struct HISTORY
{
private:
  template <typename T>
  static string NumToString( T Num, const bool CheckType = true, const int digits = 2 )
  {
    if (CheckType && (OrderType() == OP_SELL))
      Num = -Num;

    return(((Num > 0) ? "+" : "") + ((typename(T) == "double") ? ::DoubleToString(Num, digits) : (string)Num));
  }

  static string VolumeToString( double dVolume, const string Delimeter = " " )
  {
    dVolume = dVolume + 0.001;

    long Round = (long)dVolume;
    string Res = "." + ::IntegerToString((long)(dVolume * 100) % 100, 2, '0');

    Res = ((bool)(Round / 1000) ? ::IntegerToString(Round % 1000, 3, '0') : (string)(Round % 1000)) + Res;

    Round /= 1000;

    while (Round)
    {
      Res = ((bool)(Round / 1000) ? ::IntegerToString(Round % 1000, 3, '0') : (string)(Round % 1000)) + Delimeter + Res;

      Round /= 1000;
    }

    return(Res);
  }

  static string LengthToString( const datetime Length )
  {
    const int Days = (int)(Length / (24 * 3600));

    return(((Days) ? (string)Days + "d ": "") + ::TimeToString(Length, TIME_SECONDS));
  }

  static string LengthToString( long Length )
  {
    if (Length < 0)
      Length += 1000;

    return(HISTORY::LengthToString((datetime)(Length / 1000)) + "." + ::IntegerToString(Length % 1000, 3, '0'));
  }

  static int PriceToPips( const double Price, const double point )
  {
    return(point ? ((int)(Price / point + 0.1)) : 0);
  }

  static int ProfitToPips( const double PriceClose, const double PriceOpen, const double point )
  {
    return(HISTORY::PriceToPips(PriceClose, point) - HISTORY::PriceToPips(PriceOpen , point));
  }

  static string DoubleToString( const double Num, const int digits )
  {
    return(Num ? ::DoubleToString(Num, digits) : "");
  }

  static int GetDigits( double Price )
  {
    int Res = 0;

    while ((bool)(Price = ::NormalizeDouble(Price - (int)Price, 8)))
    {
      Price *= 10;

      Res++;
    }

    return(Res);
  }

  static string TypeToString( const int Type )
  {
    static const string Types[] = {"buy", "sell", "buy limit", "sell limit", "buy stop", "sell stop", "balance"};

    return(((Type < ::ArraySize(Types)) ? Types[Type] : "unknown"));
  }

  static int GetSymbolPos( const string Symb )
  {
    static string Symbols[];

    const int Size = ::ArraySize(Symbols);

    int Pos = 0;

    while ((Pos < Size) && (Symbols[Pos] != Symb))
      Pos++;

    if (Pos == Size)
    {
      ::ArrayResize(Symbols, Size + 1);

      Symbols[Pos] = Symb;
    }

    return(Pos);
  }

  //         
  static double GetTickValue()
  {
    static double TickValue[];

    const int Pos = HISTORY::GetSymbolPos(OrderSymbol());
    const int Size = ::ArraySize(TickValue);

    if (Pos >= Size)
      ::ArrayResize(TickValue, Pos + 1);

    const double TmpPips = ::NormalizeDouble(OrderClosePrice() - OrderOpenPrice(), 8); // https://www.mql5.com/ru/forum/170952/page128#comment_10696343
    const double Lots = OrderLots();

    if (TmpPips && Lots/* && OrderProfit()*/) //     - OrderType() > OP_SELL
      TickValue[Pos] = ::MathAbs(OrderProfit() / (TmpPips * Lots));

    return(TickValue[Pos]);
  }

public:
  static int Number;
  static double PrevBalance;

  TICKET_TYPE Ticket;
  string OpenTime;
  string Type;
  string Lots;
  string Symbol;
  string OpenPrice;
  string StopLoss;
  string TakeProfit;
  string CloseTime;
  string ClosePrice;
  string Commission;
  string Swap;
  string Profit;
  string Comment;
  string Pips;
  string SlipPage;
  string MagicNumber;
  string LengthTime;
  string Balance;

  HISTORY( const bool Init = false )
  {
    if (Init)
    {
      HISTORY::Number = 0;
      HISTORY::PrevBalance = 0;
    }

    static const int CurrencyDigits = (int)::AccountInfoInteger(ACCOUNT_CURRENCY_DIGITS);

    const bool IsClosed = OrderCloseTime();
    const int digits = (int)::SymbolInfoInteger(OrderSymbol(), SYMBOL_DIGITS);
    const double point = ::SymbolInfoDouble(OrderSymbol(), SYMBOL_POINT);
    const double TickValue = (OrderType() <= OP_SELL) ? HISTORY::GetTickValue() : 0;
    const int VolumeDigits = HISTORY::GetDigits(::SymbolInfoDouble(OrderSymbol(), SYMBOL_VOLUME_STEP));
    const bool SlipPage2 = ((OrderOpenPrice() != OrderOpenPriceRequest()) || (OrderClosePrice() != OrderClosePriceRequest()));

    this.Ticket = OrderTicket();
    this.OpenTime = (string)OrderOpenTime() + "." + ::IntegerToString(OrderOpenTimeMsc() % 1000, 3, '0');
    this.Type = HISTORY::TypeToString(OrderType());
    this.Lots =  HISTORY::DoubleToString(OrderLots(), VolumeDigits) +
                 ((OrderType() <= OP_SELL) ? (" (" + HISTORY::VolumeToString(OrderLots() * TickValue * (OrderOpenPrice() + OrderClosePrice())) + ")") : "");

    this.Symbol = (OrderType() == OP_BALANCE) ? "" : OrderSymbol() + (::SymbolInfoInteger(OrderSymbol(), SYMBOL_CUSTOM) ?
                                                                      " (" + ::SymbolInfoString(OrderSymbol(), SYMBOL_CURRENCY_PROFIT) + ")" : "");
    this.OpenPrice = HISTORY::DoubleToString(OrderOpenPrice(), digits);
    this.StopLoss = HISTORY::DoubleToString(OrderStopLoss(), digits);
    this.TakeProfit = HISTORY::DoubleToString(OrderTakeProfit(), digits);

    this.CloseTime = (OrderType() == OP_BALANCE) ? "" :(string)OrderCloseTime() + "." + ::IntegerToString(OrderCloseTimeMsc() % 1000, 3, '0');
    this.ClosePrice = HISTORY::DoubleToString(OrderClosePrice(), digits);
    this.Commission = HISTORY::DoubleToString(OrderCommission(), CurrencyDigits);
    this.Swap = HISTORY::DoubleToString(OrderSwap(), CurrencyDigits);

    this.Profit = ((OrderType() > OP_SELL) && (OrderType() != OP_BALANCE)) ? "" : (HISTORY::NumToString(OrderProfit(), false, CurrencyDigits) +
                  (SlipPage2 ? " (" + HISTORY::NumToString(OrderLots() * (OrderClosePriceRequest() - OrderOpenPriceRequest()) * TickValue, true, CurrencyDigits) + ")" : ""));
    this.Comment = OrderComment();

    this.Pips = (OrderType() <= OP_SELL) ? HISTORY::NumToString(HISTORY::ProfitToPips(OrderClosePrice(), OrderOpenPrice(), point)) +
                                          (SlipPage2 ? " (" + HISTORY::NumToString(HISTORY::ProfitToPips(OrderClosePriceRequest(), OrderOpenPriceRequest(), point)) + ")" : "") : "";
    this.SlipPage = ((OrderType() <= OP_SELL) && SlipPage2) ?
                    HISTORY::NumToString(HISTORY::ProfitToPips(OrderOpenPriceRequest(), OrderOpenPrice(), point)) + ", " +
                    HISTORY::NumToString(HISTORY::ProfitToPips(OrderClosePrice(), OrderClosePriceRequest(), point)) : "";
    this.MagicNumber = !OrderMagicNumber() ? "" : (string)OrderMagicNumber();

    MqlTick Tick = {0};

  #ifdef __VIRTUAL__
    if (!VIRTUAL::GetHandle() || !::SymbolInfoTick(_Symbol, Tick))
  #endif // __VIRTUAL__
    Tick.time_msc = (long)::TimeCurrent() * 1000;

    this.LengthTime = (OrderType() == OP_BALANCE) ? "" :
                      HISTORY::LengthToString((IsClosed ? OrderCloseTimeMsc() : Tick.time_msc) - OrderOpenTimeMsc());

    if (IsClosed)
    {
      HISTORY::Number++;
      HISTORY::PrevBalance += OrderProfit() + OrderSwap() + OrderCommission();
    }

    this.Balance = ::DoubleToString(PrevBalance, CurrencyDigits);
  }
};

static int HISTORY::Number = 0;
static double HISTORY::PrevBalance = 0;